/******************************************************************************
* super_tee.c - $Revision: 1.5 $ - $Date: 2005/11/30 14:59:18 $
*
*   Copyright (C) 2003,2005 Craig B. Agricola (craig@theagricolas.org)
*
* This program started as a Perl script to make it easy to capture all of the
* output from a command, but still be able to easily search for output that
* came from the standard error stream.  This is particularly useful when doing
* long builds, for instance, so that you can log everything and at the end of
* the build, quickly check to see if there were any compilation errors.
*
* Since part of the process for labelling each line with the source (either
* STDOUT or STDERR) is to buffer output for each stream until we see a
* new-line, it quickly became apparent that one of the handy benefits of this
* script is that it disentangles the output from the two streams, which often
* happens when a sub-processes is spawned for some reason, and that sub-process
* generates error messages.  Often these error messages can be embedded in
* the middle of the regular output from the parent process, which makes it
* very difficult to spot visually, and even more difficult to check for with
* automatic post-processing searches of the output.
*
* Eventually I wanted to use the super_tee.pl Perl script on a machine that
* didn't have Perl installed, and rather than go through the process of
* installing Perl, I just went ahead and re-wrote the program in C, making it
* more robust in the process.
*
* BUILDING
*
*   gcc -O3 -o super_tee super_tee.c
*
* USAGE
*
*   supertee [-tna] <outputfile> [cmdline...]
*            -t  Add time stamps to the output
*            -n  Suppress the markings indicating the source of
*                  the lines (stdout vs stderr)
*            -a  Append to the output file rather than overwriting
*
* LICENSE
*
*   super_tee is free software; you can redistribute it and/or modify it under
*   the terms of the GNU General Public License as published by the Free
*   Software Foundation; either version 2, or (at your option) any later
*   version.
*
*   super_tee is distributed in the hope that it will be useful, but WITHOUT
*   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
*   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
*   more details.
*
*   You should be able to download a copy of the GNU General Public License
*   from http://www.gnu.org/licenses/gpl.txt If not, write to the Free Software
*   Foundation, 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA.
*
******************************************************************************/

#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <sys/stat.h>
#include <time.h>

#define BUF_READ 3
#define BUF_SIZE 1024

int process_input(int input_fd, int std_fd, int output_fd,
                  char *prefix, int add_time_stamps,
                  char *buf, size_t *buf_len)
{
  int     hit_end = 0;
  int     lines_processed = 0;
  size_t  read_len;
  char   *line_start;
  char   *p;
  size_t  prefix_len;
  char   *time_stamp;
  size_t  time_stamp_len;

  prefix_len = strlen(prefix);
  if (add_time_stamps) {
    time_t now;
    time(&now);
    time_stamp = ctime(&now);
    time_stamp_len = strlen(time_stamp) - 1;
  }

  while((lines_processed == 0) && (! hit_end)) {
    /* Read in a chunk of input */
    read_len = BUF_READ;
    if (*buf_len + read_len > BUF_SIZE)
      read_len = BUF_SIZE - *buf_len;
    read_len = read(input_fd, &buf[*buf_len], read_len);
    if (read_len <= 0)
      hit_end = 1;
    else
      *buf_len += read_len;

    /* Scan forward for end of lines */
    p = line_start = buf;
    while(p != &buf[*buf_len]) {
      for(; (*p != '\n') && (p != &buf[*buf_len]); p++);
      if ((p != &buf[*buf_len]) && (*p=='\n')) {
        p++;
        if (add_time_stamps) {
          write(output_fd, time_stamp, time_stamp_len);
          write(output_fd, ": ", 2);
        }
        write(output_fd, prefix, prefix_len);
        write(output_fd, line_start, p - line_start);
        write(std_fd, line_start, p - line_start);
        line_start = p;
        lines_processed++;
      }
    }

    /* Slide the remaining data to the front of the buffer */
    if (lines_processed > 0) {
      p = buf;
      while(line_start != &buf[*buf_len])
        *p++ = *line_start++;
      *buf_len = p - buf;
    }

    /* If nothing was processed, and we are out of buffer, just make a line...*/
    else if ((*buf_len == BUF_SIZE) || (hit_end && (*buf_len > 0))) {
      if (add_time_stamps) {
        write(output_fd, time_stamp, time_stamp_len);
        write(output_fd, ": ", 2);
      }
      write(output_fd, prefix, prefix_len);
      write(output_fd, buf, *buf_len);
      write(std_fd, buf, *buf_len);
      write(output_fd, "\n", 1);
      *buf_len = 0;
      lines_processed++;
    }
  }
  return(hit_end);
}

int main(int argc, char* argv[]) {
  struct stat output_file_stat;
  int         output_fd;
  int         child_pid;
  int         stdout_pipe[2], stderr_pipe[2];
  int         all_done;
  fd_set      pipes_fd_set;
  int         pipes_fd_max;
  size_t      out_buf_len;
  char        out_buf[BUF_SIZE];
  size_t      err_buf_len;
  char        err_buf[BUF_SIZE];
  int         add_time_stamps = 0;
  int         no_source_marks = 0;
  int         append_mode = 0;

  argv++; argc--;
  while (argc && (argv[0][0]=='-')) {
    for(argv[0]++; argv[0][0]; argv[0]++) {
      switch (argv[0][0]) {
        case 't': add_time_stamps=1;    break;
        case 'n': no_source_marks=1;    break;
        case 'a': append_mode=1;        break;
        default:  argv[0][1]=0; argc=1; break;
      }
    }
    argv++; argc--;
  }

  if (argc < 2) {
    printf("Usage: supertee [-tna] <outputfile> [cmdline...]\n");
    printf("Options:\n");
    printf("         -t  Add time stamps to the output\n");
    printf("         -n  Suppress the markings indicating the source of\n");
    printf("               the lines (stdout vs stderr)\n");
    printf("         -a  Append to the output file rather than overwriting\n");
    exit(-1);
  }

  if (!stat(argv[0], &output_file_stat)) {
    if ((output_file_stat.st_mode & (S_IXUSR|S_IXGRP|S_IXOTH)) != 0) {
      fprintf(stderr, "Output file (%s) already exists and has execute " \
                      "permissions!  Did you forget to pass the name of " \
                      "the output file?\n", argv[0]);
      exit(-1);
    }
  }
  output_fd = open(argv[0], O_WRONLY|O_CREAT|(append_mode?O_APPEND:O_TRUNC),
                            S_IWUSR|S_IRUSR|S_IRGRP|S_IROTH);
  if (output_fd == -1) {
    fprintf(stderr, "Couldn't open %s for writing: %s\n", argv[0], strerror());
    exit(-1);
  }

  argv++; argc--;

  if ((pipe(stdout_pipe)==-1) || (pipe(stderr_pipe)==-1)) {
    perror("Failed to create output pipes");
    exit(-1);
  }

  if(!(child_pid=fork())) {
    dup2(stdout_pipe[1], STDOUT_FILENO);
    dup2(stderr_pipe[1], STDERR_FILENO);
    close(stdout_pipe[0]);
    close(stderr_pipe[0]);
    execvp(argv[0], argv);
    perror("Failed to exec");
    exit(-1);
  } else {
    close(STDIN_FILENO);
    close(stdout_pipe[1]);
    close(stderr_pipe[1]);

    out_buf_len = 0;
    err_buf_len = 0;

    all_done = 0;
    while(all_done != 0x3) {
      FD_ZERO(&pipes_fd_set);
      FD_SET(stdout_pipe[0], &pipes_fd_set);
      FD_SET(stderr_pipe[0], &pipes_fd_set);

      pipes_fd_max = stdout_pipe[0];
      if (stderr_pipe[0] > pipes_fd_max)
        pipes_fd_max = stderr_pipe[0];
      pipes_fd_max++;

      if(select(pipes_fd_max, &pipes_fd_set, NULL, NULL, NULL) > 0) {
        if (FD_ISSET(stdout_pipe[0], &pipes_fd_set)) {
          if (process_input(stdout_pipe[0], STDOUT_FILENO, output_fd,
                            no_source_marks?"":"STDOUT: ", add_time_stamps,
                            out_buf, &out_buf_len))
            all_done |= 0x1;
        }
        if (FD_ISSET(stderr_pipe[0], &pipes_fd_set)) {
          if (process_input(stderr_pipe[0], STDERR_FILENO, output_fd,
                            no_source_marks?"":"STDERR: ", add_time_stamps,
                            err_buf, &err_buf_len))
            all_done |= 0x2;
        }
      } else
        all_done = 0x3;
    }

    waitpid(child_pid, 0, 0);
  }

  return(0);
}

/* vim: set sw=2: */
